==> v0.6.0.tar.gz: shellfu/mkit/include/build.sh <==
#!/bin/bash

. mkit/include/ini.sh || die "cannot import ini.sh"


build() {
    #
    # Add meat to all skeletons
    #
    local srcpath
    find src -type f -name '*.skel' \
     | while read srcpath;
       do
           build1 "$srcpath"
       done
}

build1() {
    #
    # Process one skeleton
    #
    local srcpath dstpath
    srcpath=$1
    dstpath=${srcpath%.skel}
    case $dstpath in
        *.md)   cat $srcpath \
                  | expand_includes \
                  | expand_variables \
                > $dstpath ;;
        *)      cat $srcpath \
                  | expand_variables \
                > $dstpath ;;
    esac
    echo $dstpath >> built.list
}

build_manpages() {
    local manfile mdfile
    if command -v ronn >/dev/null;
    then
        ini lskeys "files:man" \
          | while read manfile;
            do
                mdfile=$manfile.md
                ronn -r $mdfile
                echo $manfile >> built.list
            done
    else
        echo "ronn is not installed"
        return 1
    fi
}

clean() {
    #
    # Clean up tree after building
    #
    test -f built.list && {
        cat built.list | xargs rm -f
        rm -f built.list
    } || :
}

dist() {
    #
    # Create distributable tarball
    #
    local version=$(get_version)
    local dirname=$MKIT_PROJNAME-$version
    mkdir -p $dirname
    local item
    cp -R $(ini values "lists:dist") $dirname
    sed -i -e "s/^VERSION = .*/VERSION = $version/" $dirname/config.mk
    tar -cf $dirname.tar $dirname
    gzip $dirname.tar
    rm -rf $dirname
}

expand_includes() {
    #
    # Expand include directives
    #
    # Expand e.g. `<!-- include4: foo.sh -->` to include code of foo.sh
    #
    perl -we '
        use strict;
        my $text;
        while (<>) {
            chomp;
            if (m/<!-- include4: (\S+) -->/) {
                open my $fh, $1 or warn "cannot find: $1";
                my $text = do { local($/); <$fh> };
                close $fh;
                $text =~ s/^(.)/    $1/gm;
                chomp $text;
                print "$text\n";
            } else {
                print "$_\n";
            }
        }
    '
}

expand_variables() {
    #
    # Expand variable values
    #
    local script=$(mktemp --tmpdir mkit-tmp.XXXXXXXXXX)
    local varname varvalue
    ini lskeys "vars" \
      | while read varname;
        do
            varvalue="$(ini 1value "vars:$varname" | sed -e 's/\$/\\$/' )"
            echo "s|$varname|$varvalue|;" >> $script
        done
    echo "s|__CODENAME__|$CODENAME|;"     >> $script
    echo "s|__VERSION__|$(get_version)|;" >> $script
    perl -wp $script
    rm $script
}

get_version() {
    #
    # Build semver version string with build metadata
    #
    # Build version string from available info using following
    # logic:
    #
    #  1. use VERSION (from config.mk)
    #  2. if we are in git, override the version with last tag
    #  3. if set, add STAGE (from config.mk) as pre-release ID
    #     (afer dash)
    #  4. if we are at a later commit than the last tag, add commit
    #     sha1 to build metadata (after plus sign)
    #  5. if the tree is "dirty", i.e. has uncommited changes,
    #     add "dirty" to build metadata
    #
    # The version is compatible with SemVer 2.0.0.
    #
    # Examples:
    #
    #     myprog v1.0.7                         # all clear
    #     myprog v1.0.7-alpha                   # STAGE="alpha"
    #     myprog v1.0.7-alpha+g1aef811          # ^^ + some commits after
    #     myprog v1.0.7-alpha+g1aef811.dirty    # ^^ + tree edited
    #     myprog v1.0.7-alpha+dirty             # tag OK but tree edited
    #     myprog v1.0.7+dirty                   # ^^ but no stage
    #
    # Note that versions with "dirty" should be perceived as kind of
    # dangerous outside developer's own machine.  Versions with sha1 are
    # safer but must not be released.
    #
    # I have considered decorating the git commit refs to make them
    # sort of sortable (e.g. "r1.g1aef811"), but on second thought,
    # I don't think it's good idea to give *any* semantics to meta-data
    # at all.  First, there is no rule that r1<r2<r3; a commit can be
    # removing what other just added and one change can be split to
    # multiple commits.  Also, the whole thing breaks anyway once you
    # rebase your branch (no, it's not a sin).  The sole purpose of
    # meta-data is to *identify* the code, and provide safe path back
    # to tree; commit refs are already perfect for that.
    #
    # FIXME:  The assumpton "if we are in git repo, we can read version
    #         from tags" fails if we are in shallow clone made from
    #         other than a tagged commit.
    #
    local version=$VERSION
    local stage=$STAGE
    if git rev-parse HEAD >&/dev/null;
    then    # we are in git repo... so we can get smart
        local lasttag=$(git tag | grep ^v | sort -V | tail -n1)
        if ! git describe --tags --exact-match HEAD >&/dev/null;
        then    # we are not at later commit than the last tag
            local sha=g$(git log -1 --pretty=format:%h HEAD)
        fi
        if test "$(git diff --shortstat 2>/dev/null)" != "";
        then    # thr tree is "dirty", i.e. has been edited
            local dirty=dirty
        fi
        version=${lasttag:1}
        local suffix=""
        case $stage:$sha:$dirty in
            ::)         suffix=""                    ;;
            ::dirty)    suffix="+$dirty"             ;;
            :g*:)       suffix="+$sha"               ;;
            :g*:dirty)  suffix="+$sha.dirty"         ;;
            *::)        suffix="-$stage"             ;;
            *::dirty)   suffix="-$stage+$dirty"      ;;
            *:g*:)      suffix="-$stage+$sha"        ;;
            *:g*:dirty) suffix="-$stage+$sha.$dirty" ;;
        esac
        version=$version$suffix
    fi
    echo $version
}

==> v0.6.1.tar.gz: shellfu/mkit/include/build.sh <==
#!/bin/bash

. mkit/include/ini.sh || die "cannot import ini.sh"


build() {
    #
    # Add meat to all skeletons
    #
    local srcpath
    find src -type f -name '*.skel' \
     | while read srcpath;
       do
           build1 "$srcpath"
       done
}

build1() {
    #
    # Process one skeleton
    #
    local srcpath dstpath
    srcpath=$1
    dstpath=${srcpath%.skel}
    case $dstpath in
        *.md)   cat $srcpath \
                  | expand_includes \
                  | expand_variables \
                > $dstpath ;;
        *)      cat $srcpath \
                  | expand_variables \
                > $dstpath ;;
    esac
    echo $dstpath >> built.list
}

build_manpages() {
    local manfile mdfile
    if command -v ronn >/dev/null;
    then
        ini lskeys "files:man" \
          | while read manfile;
            do
                mdfile=$manfile.md
                ronn -r $mdfile
                echo $manfile >> built.list
            done
    else
        echo "ronn is not installed"
        return 1
    fi
}

clean() {
    #
    # Clean up tree after building
    #
    test -f built.list && {
        cat built.list | xargs rm -f
        rm -f built.list
    } || :
}

dist() {
    #
    # Create distributable tarball
    #
    local version=$(get_version)
    local dirname=$MKIT_PROJNAME-$version
    mkdir -p $dirname
    local item
    cp -R $(ini values "lists:dist") $dirname
    sed -i -e "s/^VERSION = .*/VERSION = $version/" $dirname/config.mk
    tar -cf $dirname.tar $dirname
    gzip $dirname.tar
    rm -rf $dirname
}

expand_includes() {
    #
    # Expand include directives
    #
    # Expand e.g. `<!-- include4: foo.sh -->` to include code of foo.sh
    #
    perl -we '
        use strict;
        my $text;
        while (<>) {
            chomp;
            if (m/<!-- include4: (\S+) -->/) {
                open my $fh, $1 or warn "cannot find: $1";
                my $text = do { local($/); <$fh> };
                close $fh;
                $text =~ s/^(.)/    $1/gm;
                chomp $text;
                print "$text\n";
            } else {
                print "$_\n";
            }
        }
    '
}

expand_variables() {
    #
    # Expand variable values
    #
    local script=$(mktemp --tmpdir mkit-tmp.XXXXXXXXXX)
    local varname varvalue
    ini lskeys "vars" \
      | while read varname;
        do
            varvalue="$(ini 1value "vars:$varname" | sed -e 's/\$/\\$/' )"
            echo "s|$varname|$varvalue|;" >> $script
        done
    echo "s|__CODENAME__|$CODENAME|;"     >> $script
    echo "s|__VERSION__|$(get_version)|;" >> $script
    perl -wp $script
    rm $script
}

get_version() {
    #
    # Build semver version string with build metadata
    #
    # Build version string from available info using following
    # logic:
    #
    #  1. use VERSION (from config.mk)
    #  2. if we are in git, override the version with last tag
    #  3. if set, add STAGE (from config.mk) as pre-release ID
    #     (afer dash)
    #  4. if we are at a later commit than the last tag, add branch
    #     name and commit sha1 to build metadata (after plus sign)
    #  5. if the tree is "dirty", i.e. has uncommited changes,
    #     add "dirty" to build metadata
    #
    # The version is compatible with SemVer 2.0.0.
    #
    # Examples:
    #
    #     myprog v1.0.7                         # all clear
    #     myprog v1.0.7-alpha                   # STAGE="alpha"
    #     myprog v1.0.7-alpha+g1aef811.master   # ^^ + some commits after
    #     myprog v1.0.7-alpha+gf14fc4f.api2     # ^^ + on a feature branch
    #     myprog v1.0.7-alpha+gf14fc4f.api2.dirty  # ^^ + tree edited
    #     myprog v1.0.7-alpha+dirty             # tag OK but tree edited
    #     myprog v1.0.7+dirty                   # ^^ but no stage
    #
    # Note that versions with "dirty" should be perceived as kind of
    # dangerous outside developer's own machine.  Versions with sha1 are
    # safer but must not be released.
    #
    # I have considered decorating the git commit refs to make them
    # sort of sortable (e.g. "r1.g1aef811"), but on second thought,
    # I don't think it's good idea to give *any* semantics to meta-data
    # at all.  First, there is no rule that r1<r2<r3; a commit can be
    # removing what other just added and one change can be split to
    # multiple commits.  Also, the whole thing breaks anyway once you
    # rebase your branch (no, it's not a sin).  The sole purpose of
    # meta-data is to *identify* the code, and provide safe path back
    # to tree; commit refs are already perfect for that.
    #
    # FIXME:  The assumpton "if we are in git repo, we can read version
    #         from tags" fails if we are in shallow clone made from
    #         other than a tagged commit.
    #
    local version=$VERSION
    local stage=$STAGE
    grep ":" <<<"$stage" && warn "colon in STAGE may corrupt version data: $stage"
    if git rev-parse HEAD >&/dev/null;
    then    # we are in git repo... so we can get smart
        local lasttag=$(git tag | grep ^v | sort -V | tail -n1)
        if ! git describe --tags --exact-match HEAD >&/dev/null;
        then    # we are at a later commit than the last tag
            local sha=g$(git log -1 --pretty=format:%h HEAD)
            local curbranch=$(git rev-parse --abbrev-ref HEAD)
            local commit="$curbranch.$sha"
        fi
        if test "$(git diff --shortstat 2>/dev/null)" != "";
        then    # the tree is "dirty", i.e. has been edited
            local dirty=dirty
        fi
        version=${lasttag:1}
        local suffix=""
        case $stage:$commit:$dirty in
            ::)        suffix=""                       ;;
            ::dirty)   suffix="+$dirty"                ;;
            :*:)       suffix="+$commit"               ;;
            :*:dirty)  suffix="+$commit"               ;;
            *::)       suffix="-$stage"                ;;
            *::dirty)  suffix="-$stage+$dirty"         ;;
            *:*:)      suffix="-$stage+$commit"        ;;
            *:*:dirty) suffix="-$stage+$commit.$dirty" ;;
        esac
        version=$version$suffix
    fi
    echo $version
}
